import torch
import torch.autograd as autograd
from torch.autograd import Variable
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.nn.parameter import Parameter

import numpy as np

class dSprites_Model(nn.Module):

  def __init__(self, _args, _encoder, _decoder):
    super(dSprites_Model, self).__init__()

    self.dtype = 'float32'
    self.args = _args

    self.NLLloss = torch.nn.NLLLoss(reduction = 'none')
    self.CEloss =  torch.nn.CrossEntropyLoss(reduction = 'none')
    self.BCEloss = torch.nn.functional.binary_cross_entropy_with_logits

    self.encoder = _encoder
    self.decoder = _decoder

    self.decompose_hidden2style1_mu = nn.Linear(self.encoder.z_dim * 2, self.args['STYLE_LEN'], bias = True)
    self.decompose_hidden2style2_mu = nn.Linear(self.encoder.z_dim * 2, self.args['STYLE_LEN'], bias = True)
    self.decompose_hidden2content_mu = nn.Linear(self.encoder.z_dim * 2, self.args['CONTENT_LEN'], bias = True)

    self.decompose_hidden2style1_std = nn.Linear(self.encoder.z_dim * 2, self.args['STYLE_LEN'], bias = True)
    self.decompose_hidden2style2_std = nn.Linear(self.encoder.z_dim * 2, self.args['STYLE_LEN'], bias = True)
    self.decompose_hidden2content_std = nn.Linear(self.encoder.z_dim * 2, self.args['CONTENT_LEN'], bias = True)

    self.tanh = nn.Tanh()
    self.softmax = nn.Softmax(dim = -1)

    #self.adapt_unit_style1_2h = nn.Linear(self.args['STYLE_LEN'], self.encoder.z_dim, bias = True)
    #self.adapt_unit_style2_2h = nn.Linear(self.args['STYLE_LEN'], self.encoder.z_dim, bias = True)
    #self.adapt_unit_content2h = nn.Linear(self.args['CONTENT_LEN'], self.encoder.z_dim, bias = True)
    
    self.style1_mu = Parameter(torch.rand((self.args['STYLE1_LABELNUM'], self.args['STYLE_LEN'])) - 0.5)
    self.style1_logvar = Parameter(torch.rand((self.args['STYLE1_LABELNUM'], self.args['STYLE_LEN'])) - 0.5)

    self.style2_mu = Parameter(torch.rand((self.args['STYLE2_LABELNUM'], self.args['STYLE_LEN'])) - 0.5)
    self.style2_logvar = Parameter(torch.rand((self.args['STYLE2_LABELNUM'], self.args['STYLE_LEN'])) - 0.5)

    self.style1_classifier = nn.Linear(self.args['STYLE_LEN'] * 2, self.args['STYLE1_LABELNUM'], bias = True)
    self.style2_classifier = nn.Linear(self.args['STYLE_LEN'] * 2, self.args['STYLE2_LABELNUM'], bias = True)

    self.labelmu1 = Parameter(torch.rand(self.args['STYLE1_LABELNUM'], 1, self.args['CONTENT_LEN']))
    self.labellogvar1 = Parameter(torch.rand(self.args['STYLE1_LABELNUM'], 1, self.args['CONTENT_LEN']))

    self.labelmu2 = Parameter(torch.rand(self.args['STYLE2_LABELNUM'], 1, self.args['CONTENT_LEN']))
    self.labellogvar2 = Parameter(torch.rand(self.args['STYLE2_LABELNUM'], 1, self.args['CONTENT_LEN']))

  def _interpolate(self, _values, _progress):
    return _values[0] * (1.0 - _progress) + _values[1] * (_progress)

  def _sample_z(self, mu, log_var):
    eps = torch.randn_like(log_var)

    return mu + torch.einsum('ba,ba->ba', torch.exp(log_var/2),eps)

  def _sample_z_weighted(self, mus, log_vars, weights, batch_size):
    mus_repeat = mus.expand((batch_size,) + mus.shape)
    log_vars_repeat = log_vars.expand((batch_size,) + log_vars.shape)
    weights_repeat = weights.unsqueeze(-1).expand((weights.shape + (self.args['STYLE_LEN'], )))
    eps = torch.randn_like(mus_repeat)

    samples = mus_repeat + torch.exp(log_vars_repeat / 2.0) * eps
    r = samples * weights_repeat

    return r.sum(dim=1)

  def _VAE_sample_z(self, mu, log_var):
    std = torch.exp(log_var*0.5)
    eps = torch.randn_like(std)
    z = mu + std * eps

    return z

  def encode(self, x, v, l, _progress = 0.0):
    self.encoderInputs = x.unsqueeze(1)
    self.classifyLabels1 = l[:, 0]
    self.classifyLabels2 = l[:, 1]
    self.batch_size = x.shape[0]

    en_state = self.tanh(self.encoder(self.encoderInputs))
    VAE_style1_emb_mu, VAE_style1_emb_logvar, VAE_style2_emb_mu, VAE_style2_emb_logvar, VAE_content_emb_mu, VAE_content_emb_logvar = self._decompose(en_state)

    VAE_KL_style1 = torch.mean(0.5 * (torch.exp(VAE_style1_emb_logvar) + VAE_style1_emb_mu ** 2 - 1 - VAE_style1_emb_logvar))
    VAE_KL_style2 = torch.mean(0.5 * (torch.exp(VAE_style2_emb_logvar) + VAE_style2_emb_mu ** 2 - 1 - VAE_style2_emb_logvar))
    VAE_KL_content = torch.mean(0.5 * (torch.exp(VAE_content_emb_logvar) + VAE_content_emb_mu ** 2 - 1 - VAE_content_emb_logvar))

    style1_emb = self._VAE_sample_z(VAE_style1_emb_mu, VAE_style1_emb_logvar)
    style2_emb = self._VAE_sample_z(VAE_style2_emb_mu, VAE_style2_emb_logvar)
    content_emb = self._VAE_sample_z(VAE_content_emb_mu, VAE_content_emb_logvar)

    """self.style1_emb = style1_emb
    self.style2_emb = style2_emb
    self.content_emb = torch.zeros_like(content_emb)"""

    batch_style1_mu = self.style1_mu.expand((self.batch_size,) + self.style1_mu.shape)
    batch_style1_var = torch.exp(self.style1_logvar).expand((self.batch_size,) + self.style1_logvar.shape)
    style1_emb_per_dist = style1_emb.expand((self.args['STYLE1_LABELNUM'], ) + style1_emb.shape).transpose(0,1)
    prob_style1_styleemb = torch.mean(torch.exp(-(style1_emb_per_dist - batch_style1_mu)**2/(2 * batch_style1_var + 0.0000001)) / torch.sqrt(2 * np.pi * batch_style1_var), dim=-1)
    style1_z = self._sample_z_weighted(self.style1_mu, self.style1_logvar, torch.softmax(prob_style1_styleemb, dim=-1), self.batch_size)

    batch_style2_mu = self.style1_mu.expand((self.batch_size,) + self.style2_mu.shape)
    batch_style2_var = torch.exp(self.style2_logvar).expand((self.batch_size,) + self.style2_logvar.shape)
    style2_emb_per_dist = style2_emb.expand((self.args['STYLE2_LABELNUM'], ) + style2_emb.shape).transpose(0,1)
    prob_style2_styleemb = torch.mean(torch.exp(-(style2_emb_per_dist - batch_style2_mu)**2/(2 * batch_style2_var + 0.0000001)) / torch.sqrt(2 * np.pi * batch_style2_var), dim=-1)
    style2_z = self._sample_z_weighted(self.style2_mu, self.style2_logvar, torch.softmax(prob_style2_styleemb, dim=-1), self.batch_size)

    # what if here I do weighted sum instead?
    self.batch_mu1s = torch.index_select(self.style1_mu, 0, self.classifyLabels1)
    self.batch_var1s = torch.index_select(torch.exp(self.style1_logvar), 0, self.classifyLabels1)
    self.batch_mu2s = torch.index_select(self.style2_mu, 0, self.classifyLabels2)
    self.batch_var2s = torch.index_select(torch.exp(self.style2_logvar), 0, self.classifyLabels2)

    style1_z = self._sample_z(self.batch_mu1s, torch.log(self.batch_var1s)) * (1.0 - _progress / 2.0) + style1_z
    style1_z = style1_z.view(self.batch_size, self.args['STYLE_LEN'])

    style2_z = self._sample_z(self.batch_mu2s, torch.log(self.batch_var2s)) * (1.0 - _progress / 2.0) + style2_z
    style2_z = style2_z.view(self.batch_size, self.args['STYLE_LEN'])

    emb = torch.cat((style1_emb, style2_emb, content_emb), -1) 
    h = self._recombine(style1_z, style2_z, content_emb)
    scloss, prob_nll_loss = self._SC_disentangle_loss(content_emb)
    VAE_kl = VAE_KL_style1 + VAE_KL_style2 + VAE_KL_content

    style_1_sample = style1_z
    style_2_sample = style2_z
    #style_1_sample = self._sample_z(self.batch_mu1s, torch.log(self.batch_var1s))
    #style_2_sample = self._sample_z(self.batch_mu2s, torch.log(self.batch_var2s))
    
    style_combined_sample = torch.cat((style_1_sample, style_2_sample), -1)

    y1 = self.style1_classifier(style_combined_sample)
    classify_loss1 = torch.mean(self.CEloss(y1, self.classifyLabels1))
    y2 = self.style2_classifier(style_combined_sample)
    classify_loss2 = torch.mean(self.CEloss(y2, self.classifyLabels2))

    return emb, h, scloss, prob_nll_loss, VAE_kl, prob_style1_styleemb, prob_style2_styleemb, classify_loss1, classify_loss2

  def forward(self, x, v, l, _progress = 0.0):
    _, h, scloss, prob_nll_loss, VAE_kl, prob_style1_styleemb, prob_style2_styleemb, classify_loss1, classify_loss2 = self.encode(x, v, l, _progress)

    de_input_state = h
    de_outputs = self.decoder(de_input_state)

    recon_loss = self.BCEloss(de_outputs, self.encoderInputs, reduction='none')
    recon_loss = torch.squeeze(recon_loss)
    recon_loss_mean = torch.mean(recon_loss)

    recon_loss_mean = recon_loss_mean * self._interpolate(self.args['M_LOSS_RECON'], _progress) + self._interpolate(self.args['M_LOSS_BETA'], _progress) * VAE_kl
    recon_loss_mean += scloss * self._interpolate(self.args['M_LOSS_SC'], _progress)
    recon_loss_mean += prob_nll_loss * self._interpolate(self.args['M_LOSS_NLL'], _progress)

    pl = 0
    cl = 0
    #prob_style1_styleemb_mean = torch.mean(self.CEloss(prob_style1_styleemb, self.classifyLabels1))
    #prob_style2_styleemb_mean = torch.mean(self.CEloss(prob_style2_styleemb, self.classifyLabels2))
    prob_style1_styleemb_mean = torch.mean(self.NLLloss(torch.log(prob_style1_styleemb + 0.0000001), self.classifyLabels1))
    prob_style2_styleemb_mean = torch.mean(self.NLLloss(torch.log(prob_style2_styleemb + 0.0000001), self.classifyLabels2))

    #mask_style_1 =  1 - torch.nn.functional.one_hot(self.classifyLabels1, num_labels=9)
    #style1_style2_labels_disentanglement_prob = 1.0 - prob_style1_styleemb * prob_style2_styleemb
    #style1_style2_labels_disentanglement_loss = -torch.mean(torch.sum(mask_style_1 * torch.log(style1_style2_labels_disentanglement_prob + 0.0000001), dim=-1), dim=-1)
    style1_style2_labels_disentanglement_loss = 0

    KL_loss1 = torch.sum(0.5 * (torch.exp(self.style1_logvar) + self.style1_mu ** 2 - 1 - self.style1_logvar))
    KL_loss2 = torch.sum(0.5 * (torch.exp(self.style2_logvar) + self.style2_mu ** 2 - 1 - self.style2_logvar))

    loss_mean = recon_loss_mean \
              + self._interpolate(self.args['M_LOSS_KL'], _progress) * (KL_loss1 + KL_loss2) \
              + self._interpolate(self.args['M_LOSS_EMB'], _progress) * (prob_style1_styleemb_mean + prob_style2_styleemb_mean) \
              + self._interpolate(self.args['M_LOSS_CL'], _progress) * (classify_loss1 + classify_loss2) \
              + self._interpolate(self.args['M_LOSS_SD'], _progress) * (self._styletype_dissolve(self.style1_mu, self.style1_logvar, self.args['STYLE1_LABELNUM']) + self._styletype_dissolve(self.style2_mu, self.style2_logvar, self.args['STYLE2_LABELNUM'])) \
              + style1_style2_labels_disentanglement_loss * 0.01

    pl = prob_style1_styleemb_mean + prob_style2_styleemb_mean
    cl = classify_loss1 + classify_loss2

    return loss_mean, pl, cl, scloss, prob_nll_loss, de_outputs

  def _styletype_dissolve(self, style_mu, style_logvar, labelnum):
    labelmu1 = style_mu[0, :]
    labellogvar1 = style_logvar[0, :]
    labelmu1_repeat = labelmu1.repeat(labelnum, 1)
    labellogvar1_repeat = labellogvar1.repeat(labelnum, 1)

    styletypedissolve_loss = 0.5*(torch.exp(labellogvar1_repeat - style_logvar) + (labelmu1_repeat - style_mu)**2/torch.exp(style_logvar) - 1 + style_logvar - labellogvar1_repeat)
    styletypedissolve_loss = torch.mean(styletypedissolve_loss, dim = 1)

    return torch.mean(styletypedissolve_loss)

  def _SC_disentangle_loss(self, content_emb):
    style1_emb_per_dist = torch.unsqueeze(content_emb, 0).repeat(self.args['STYLE1_LABELNUM'], 1, 1, 1)
    style2_emb_per_dist = torch.unsqueeze(content_emb, 0).repeat(self.args['STYLE2_LABELNUM'], 1, 1, 1)

    batch_c_mu1 = torch.unsqueeze(self.labelmu1, 2)
    batch_c_sigma1 = torch.unsqueeze(torch.exp(self.labellogvar1), 2)
    P_c_t1 = torch.exp(-(style1_emb_per_dist - batch_c_mu1) ** 2 / (2 * batch_c_sigma1 + 0.0000001)) / torch.sqrt(2 * np.pi * batch_c_sigma1)
    P_c_t1 = torch.log(P_c_t1 + 0.0000001)
    P_c_t1 = torch.mean(P_c_t1, dim = 3)
    P_c_t1 = torch.mean(P_c_t1, dim = 1)
    P_c_t1 = torch.transpose(P_c_t1, 0,1).contiguous()
    prob_style1_contentemb_mean = torch.mean(self.NLLloss(P_c_t1, self.classifyLabels1))

    batch_c_mu2 = torch.unsqueeze(self.labelmu2, 2)
    batch_c_sigma2 = torch.unsqueeze(torch.exp(self.labellogvar2), 2)
    P_c_t2 = torch.exp(-(style2_emb_per_dist - batch_c_mu2) ** 2 / (2 * batch_c_sigma2 + 0.0000001)) / torch.sqrt(2 * np.pi * batch_c_sigma2)
    P_c_t2 = torch.log(P_c_t2 + 0.0000001)
    P_c_t2 = torch.mean(P_c_t2, dim=3)
    P_c_t2 = torch.mean(P_c_t2, dim=1)
    P_c_t2 = torch.transpose(P_c_t2, 0, 1).contiguous()
    prob_style2_contentemb_mean = torch.mean(self.NLLloss(P_c_t2, self.classifyLabels2))

    labelmu1_0 = self.labelmu1[0, :, :]
    labellogvar1_0 = self.labellogvar1[0, :, :]
    labelmu1_0_repeat = labelmu1_0.repeat(self.args['STYLE1_LABELNUM'], 1, 1)
    labellogvar1_0_repeat = labellogvar1_0.repeat(self.args['STYLE1_LABELNUM'], 1, 1)

    labelmu2_0 = self.labelmu2[0, :, :]
    labellogvar2_0 = self.labellogvar2[0, :, :]
    labelmu2_0_repeat = labelmu2_0.repeat(self.args['STYLE2_LABELNUM'], 1, 1)
    labellogvar2_0_repeat = labellogvar2_0.repeat(self.args['STYLE2_LABELNUM'], 1, 1)

    self.s1c_loss = 0.5 * (torch.exp(labellogvar1_0_repeat - self.labellogvar1) + (labelmu1_0_repeat - self.labelmu1) ** 2 \
                  / torch.exp(self.labellogvar1) - 1 + self.labellogvar1 - labellogvar1_0_repeat)
    self.s1c_loss = torch.sum(self.s1c_loss, dim=2)
    self.s1c_loss = torch.sum(self.s1c_loss, dim=1)

    self.s2c_loss = 0.5 * (torch.exp(labellogvar2_0_repeat - self.labellogvar2) + (labelmu2_0_repeat - self.labelmu2) ** 2 \
                  / torch.exp(self.labellogvar2) - 1 + self.labellogvar2 - labellogvar2_0_repeat)
    self.s2c_loss = torch.sum(self.s2c_loss, dim=2)
    self.s2c_loss = torch.sum(self.s2c_loss, dim=1)

    return torch.mean(self.s1c_loss) + torch.mean(self.s2c_loss), prob_style1_contentemb_mean + prob_style2_contentemb_mean

  def predict(self, x):
    self.encoderInputs = x.unsqueeze(1)
    self.batch_size = x.shape[0]

    en_state = self.encoder(self.encoderInputs)
    VAE_style1_emb_mu, VAE_style1_emb_logvar, VAE_style2_emb_mu, VAE_style2_emb_logvar, VAE_content_emb_mu, VAE_content_emb_logvar = self._decompose(en_state)

    style1_emb = self._VAE_sample_z(VAE_style1_emb_mu, VAE_style1_emb_logvar, self.batch_size)
    style2_emb = self._VAE_sample_z(VAE_style2_emb_mu, VAE_style2_emb_logvar, self.batch_size)
    content_emb = self._VAE_sample_z(VAE_content_emb_mu, VAE_content_emb_logvar, self.batch_size)
    
    label1_s1 = self.style1_classifier(style1_emb)
    label2_s1 = self.style1_classifier(style2_emb)
    #label_c_s1 = self.style1_classifier(content_emb)

    label1_s2 = self.style2_classifier(style1_emb)
    label2_s2 = self.style2_classifier(style2_emb)
    #label_c_s2 = self.style2_classifier(content_emb)

    return torch.argmax(label1_s1, dim = -1).detach().item(), torch.argmax(label2_s2, dim = -1).detach().item()

  def extract_useful(self, x):
    self.encoderInputs = x['enc_input']
    self.encoder_lengths = x['enc_len']

    self.batch_size = self.encoderInputs.size()[0]
    self.enc_len = self.encoderInputs.size()[1]

    en_state = self.encoder(self.encoderInputs, self.encoder_lengths)
    VAE_style1_emb_mu, VAE_style1_emb_logvar, VAE_style2_emb_mu, VAE_style2_emb_logvar, VAE_content_emb_mu, VAE_content_emb_logvar = self.decompose(en_state)

    style1_emb = self._VAE_sample_z(VAE_style1_emb_mu, VAE_style1_emb_logvar, self.batch_size)
    style2_emb = self._VAE_sample_z(VAE_style2_emb_mu, VAE_style2_emb_logvar, self.batch_size)
    content_emb = self._VAE_sample_z(VAE_content_emb_mu, VAE_content_emb_logvar, self.batch_size)

    return style1_emb, style2_emb,  content_emb

  def sample_action(self, pdf):
    pdf = pdf.cpu().detach().data
    pdf = pdf.cuda(torch.float64)
    pdf = pdf / torch.sum(pdf, dim = 1, keepdim = True)
    batch_actions = []
    for b in range(pdf.size()[0]):
      rn = np.random.multinomial(1, pdf[b,:], size = 5)
      batch_actions.append(rn)
    batch_actions = torch.tensor(batch_actions) # batch 5 attsize
    batch_actions = torch.transpose(batch_actions, 0, 1)

    return batch_actions.cuda(torch.float32)

  def _decompose(self, en_state):
    en_hidden = en_state
    s1_mu = self.decompose_hidden2style1_mu(en_hidden)
    s2_mu = self.decompose_hidden2style2_mu(en_hidden)
    c_mu = self.decompose_hidden2content_mu(en_hidden)

    s1_logvar = self.decompose_hidden2style1_std(en_hidden)
    s2_logvar = self.decompose_hidden2style2_std(en_hidden)
    c_logvar = self.decompose_hidden2content_std(en_hidden)

    return self.tanh(s1_mu), self.tanh(s1_logvar), self.tanh(s2_mu), self.tanh(s2_logvar), self.tanh(c_mu), self.tanh(c_logvar)

  def _recombine(self, style1_emb, style2_emb, content_emb):
    h = torch.cat((style1_emb, style2_emb, content_emb), -1)
    
    return h

    
