import numpy as np
import torch
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import random

import board
import train14 as dsprites

def variable_traversal(device, model, i, steps=10, start=-2, end=2):
	model.eval()

	with torch.no_grad():
		v = torch.linspace(start, end, 16)
		z = torch.tile(torch.randn((model.z_dim)), (steps, 1))
		z[:, i] = v

		img_recon = torch.sigmoid(model.decoder(z.to(device)))

		_, axs = plt.subplots(1, steps)
		axs = axs.flatten()
		for img, ax in zip(img_recon.cpu(), axs):
				ax.imshow(img.squeeze())
		plt.show()

if __name__ == "__main__":
  log = board.RunLog('r')
  log.fullname = 'r210815_1145_full_woring_not_extreme'
  

  log['EPOCHS'] = 300
  log['LR'] = 1e-3
  log['ZDIM'] = 20
  log['S_MUL'] = 0.1
  log['S_SZ'] = 250
  log['S_BATCH'] = 256
  log['STYLE_LEN'] = 1
  log['CONTENT_LEN'] = log['ZDIM'] - log['STYLE_LEN'] * 2

  log['M_LOSS_RECON'] = (10, 80)
  log['M_LOSS_BETA'] = (0.1, 0.5)
  log['M_LOSS_SC'] = (80, 80)
  log['M_LOSS_NLL'] = (100, 100)
  log['M_LOSS_KL'] = (0.1, 0.5)
  log['M_LOSS_EMB'] = (0.003, 10.0)
  log['M_LOSS_CL'] = (0.001, 5.0)
  log['M_LOSS_SD'] = (0.01, 0.01)

  data = np.load('data/dsprites_ap_full.npz', encoding='bytes', mmap_mode='r', allow_pickle=True)
  log['STYLE1_LABELNUM'] = len(data['labels_used'][0])
  log['STYLE2_LABELNUM'] = len(data['labels_used'][1])
  
  encoder = dsprites.Encoder(z_dim = log['ZDIM'])
  decoder = dsprites.Decoder(z_dim = log['ZDIM'])
  model = dsprites.dSprites_Model(log, encoder, decoder)
  model.load_state_dict(torch.load(dsprites.DEFAULT_DIR + '/' + log.fullname + dsprites.DEFAULT_EXT))
  model.eval()

  with torch.no_grad():
    for i in range(32):
      n = i
      img = data['imgs'][n]
      val = data['latents_values'][n]
      lab = data['labels'][n]
      y = []
      r = model.encode(torch.from_numpy(img).float().unsqueeze(0).expand((512,) + img.shape),
        torch.from_numpy(val).unsqueeze(0).expand((512,) + val.shape),
        torch.from_numpy(lab - min(data['labels_used'][0])).unsqueeze(0).expand((512,) + lab.shape),
        1.0)

      print('i:',i)
      emb = r[0].mean(dim=0)[1]
      print(emb)
      print(data['latents_values'][n])
